# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later


class EndOfConnection(Exception):
    pass


class NonFatalSSLError(Exception):
    pass


class FallbackLanguageError(Exception):
    pass


class StanzaDecrypted(Exception):
    pass



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Presence
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class PGPLegacy(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pgplegacy_message,
                ns=Namespace.ENCRYPTED,
                priority=7,
            ),
            StanzaHandler(
                name="presence",
                callback=self._process_signed,
                ns=Namespace.SIGNED,
                priority=15,
            ),
        ]

    @staticmethod
    def _process_signed(
        _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        signed = stanza.getTag("x", namespace=Namespace.SIGNED)
        if signed is None:
            return

        properties.signed = signed.getData()

    def _process_pgplegacy_message(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        pgplegacy = stanza.getTag("x", namespace=Namespace.ENCRYPTED)
        if pgplegacy is None:
            self._log.warning("No x node found")
            self._log.warning(stanza)
            return

        data = pgplegacy.getData()
        if not data:
            self._log.warning("No data in x node found")
            self._log.warning(stanza)
            return

        self._log.info("Encrypted message received")
        properties.pgp_legacy = data



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import ParamSpec
from typing import TYPE_CHECKING
from typing import TypeVar

import inspect
import logging
import weakref
from collections.abc import Callable
from collections.abc import Generator
from enum import IntEnum
from functools import wraps

from gi.repository import GLib

from nbxmpp.errors import BaseError
from nbxmpp.errors import CancelledError
from nbxmpp.errors import is_error
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.errors import TimeoutStanzaError
from nbxmpp.modules.util import make_func_arguments_string
from nbxmpp.simplexml import Node

if TYPE_CHECKING:
    from nbxmpp.client import Client
    from nbxmpp.dispatcher import NBXMPPModuleT

log = logging.getLogger("nbxmpp.task")


class _ResultSet:
    pass


ResultSet = _ResultSet()

T = TypeVar("T")
P = ParamSpec("P")


class NoType:
    pass


class TaskState(IntEnum):
    INIT = 0
    RUNNING = 1
    FINISHED = 2
    CANCELLED = 3

    @property
    def is_init(self):
        return self == TaskState.INIT

    @property
    def is_running(self):
        return self == TaskState.RUNNING

    @property
    def is_finished(self):
        return self == TaskState.FINISHED

    @property
    def is_cancelled(self):
        return self == TaskState.CANCELLED


def _setup_task(
    task: Task, client: Client, callback: Callable[..., Any] | None, user_data: Any
) -> Task:
    client.add_task(task)
    task.set_finalize_func(client.remove_task)
    task.set_user_data(user_data)
    if callback is not None:
        task.add_done_callback(callback)
    task.start()
    return task


def iq_request_task(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def func_wrapper(
        self: NBXMPPModuleT,
        *args: Any,
        timeout: int | None = None,
        callback: Callable[..., Any] | None = None,
        user_data: Any = None,
        **kwargs: Any,
    ) -> T:
        if self._log.isEnabledFor(logging.INFO):  # type: ignore
            self._log.info(make_func_arguments_string(func, self, args, kwargs))  # type: ignore
        task = IqRequestTask(
            func(self, *args, **kwargs), self._log, self._client  # type: ignore
        )
        task.set_timeout(timeout)
        return _setup_task(task, self._client, callback, user_data)  # type: ignore

    return func_wrapper


def is_fatal_error(error: BaseError | Exception | Any) -> bool:
    if is_error(error):
        assert isinstance(error, BaseError)
        return error.is_fatal
    return isinstance(error, Exception)


class Task:
    """
    Base class for wrapping a generator method.

    It runs the generator depending on what the generator yields. If the
    generator yields another generator a sub task is created. If it yields
    a type defined in _process_types, _run_async() is called which needs to
    be implemented by classes.

    the implementation of _run_async() must be really async, means it should
    not call _async_finished() in the same mainloop cycle. Otherwise sub tasks
    may break. _async_finished() needs to call _next_step(result).
    """

    _process_types = (NoType,)

    def __init__(
        self, gen: Generator[Any, Any, Any], logger: logging.Logger = log
    ) -> None:
        self._logger = logger
        self._gen = gen
        self._done_callbacks: list[Callable[..., Any]] = []
        self._sub_task: Task | None = None
        self._result = None
        self._error = None
        self._user_data: Any | None = None
        self._timeout: int | None = None
        self._timeout_id: int | None = None
        self._finalize_func: Callable[..., Any] | None = None
        self._finalize_context: Any = None
        self._state = TaskState.INIT

    @property
    def state(self) -> TaskState:
        return self._state

    def add_done_callback(
        self, callback: Callable[..., Any], weak: bool = True
    ) -> None:
        if self._state.is_finished or self._state.is_cancelled:
            raise RuntimeError("Task is finished")

        if weak:
            if inspect.ismethod(callback):
                callback = weakref.WeakMethod(callback)
            elif inspect.isfunction(callback):
                callback = weakref.ref(callback)
            else:
                raise ValueError("Unknown callback object")

        self._done_callbacks.append(callback)

    def set_timeout(self, timeout: int | None) -> None:
        self._timeout = timeout

    def start(self) -> None:
        if not self._state.is_init:
            raise RuntimeError("Task already started")

        if self._timeout is not None:
            self._logger.info(
                "Add timeout for task: %s s, task id: %s", self._timeout, id(self)
            )
            self._timeout_id = GLib.timeout_add_seconds(self._timeout, self._on_timeout)

        self._state = TaskState.RUNNING
        next(self._gen)
        self._next_step(self)

    def _run_async(self, data: Any) -> None:
        raise NotImplementedError

    def _async_finished(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError

    def _sub_task_completed(self, task: Task) -> None:
        self._sub_task = None
        if not self._state.is_running:
            return

        result = task.get_result()
        if is_fatal_error(result):
            self._error = result
            self._set_finished()
        else:
            self._next_step(result)

    def _next_step(self, result: Any) -> None:
        try:
            res = self._gen.send(result)
        except StopIteration:
            self._set_finished()
            return

        except Exception as error:
            self._log_if_fatal(error)
            self._error = error
            self._set_finished()
            return

        if isinstance(res, self._process_types):
            self._run_async(res)

        elif isinstance(res, Task):
            if self._sub_task is not None:
                raise RuntimeError("Only one sub task can be active")

            self._sub_task = res
            self._sub_task.add_done_callback(self._sub_task_completed, weak=False)

        else:
            if res is not ResultSet:
                self._result = res
            self._set_finished()

    def _set_finished(self) -> None:
        self._state = TaskState.FINISHED
        self._invoke_callbacks()
        self._finalize()

    def _log_if_fatal(self, error: BaseError | Exception | Any) -> None:
        if is_error(error):
            assert isinstance(
                error,
                StanzaError
                | MalformedStanzaError
                | TimeoutStanzaError
                | CancelledError,
            )
            if error.is_fatal:
                self._logger.log(error.log_level, error)

        elif isinstance(error, Exception):
            self._logger.exception("Fatal Exception")

    def _invoke_callbacks(self) -> None:
        for callback in self._done_callbacks:
            if isinstance(callback, weakref.WeakMethod | weakref.ref):
                callback = callback()  # type: ignore
                if callback is None:
                    return

            # Be conservative with catching exceptions here
            # For example unittests raise Assertion errors
            # which should not be catched here
            try:
                callback(self)
            except CancelledError:
                pass

    def set_result(self, result: Any):
        self._result = result
        return ResultSet

    def get_result(self) -> Any:
        # if self._error is None, there was no error
        # but None is a valid value for self._result
        if self._error is not None:
            return self._error
        return self._result

    def finish(self) -> Any:
        if self._error is not None:
            raise self._error
        return self._result

    def set_user_data(self, user_data: Any) -> None:
        self._user_data = user_data

    def get_user_data(self) -> Any:
        return self._user_data

    def set_finalize_func(self, func: Callable[..., Any], context: Any = None) -> None:
        self._finalize_func = func
        self._finalize_context = context

    def _on_timeout(self) -> None:
        self._logger.info("Timeout reached, task id: %s", id(self))
        if not self._state.is_running:
            return

        self._timeout_id = None

        if self._sub_task is not None:
            self._sub_task.cancel(invoke_callbacks=False)

        self._error = TimeoutStanzaError()
        self._set_finished()

    def cancel(self, invoke_callbacks: bool = True) -> None:
        if not self._state.is_running:
            return

        self._state = TaskState.CANCELLED
        if self._sub_task is not None:
            self._sub_task.cancel(invoke_callbacks=False)

        self._error = CancelledError()
        if invoke_callbacks:
            self._invoke_callbacks()
        self._finalize()

    def _finalize(self) -> None:
        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None
        self._done_callbacks.clear()
        self._sub_task = None
        self._error = None
        self._result = None
        self._user_data = None
        self._gen.close()
        if self._finalize_func is not None:
            self._finalize_func(self, self._finalize_context)


class IqRequestTask(Task):
    """
    A Task for running IQ requests

    """

    _process_types = (Node,)

    def __init__(
        self, gen: Generator[Any, Any, Any], logger: logging.Logger, client: Client
    ) -> None:
        super().__init__(gen, logger)
        self._client = client
        self._iq_id: str | None = None

    def _run_async(self, data: Node) -> None:
        assert self._client is not None
        self._iq_id = self._client.send_stanza(
            data, callback=self._async_finished, timeout=self._timeout
        )

    def _async_finished(
        self, _client: Client, result: Any, *args: Any, **kwargs: Any
    ) -> None:
        if self._state == TaskState.CANCELLED:
            return

        if result is None:
            self._error = TimeoutStanzaError()
            self._set_finished()
            return

        self._next_step(result)

    def _finalize(self) -> None:
        if self._iq_id is not None:
            self._client._dispatcher.remove_iq_callback(self._iq_id)  # type: ignore
        self._client = None
        super()._finalize()



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Any
from typing import Literal
from typing import NamedTuple
from typing import TYPE_CHECKING

from dataclasses import dataclass

from nbxmpp.const import MessageType
from nbxmpp.errors import is_error
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import PubSubStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.util import finalize
from nbxmpp.modules.util import process_response
from nbxmpp.modules.util import raise_if_error
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.structs import CommonResult
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import PubSubEventData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class PubSub(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_base,
                ns=Namespace.PUBSUB_EVENT,
                priority=15,
            ),
        ]

    def _process_pubsub_base(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        if properties.type not in (MessageType.HEADLINE, MessageType.NORMAL):
            return

        properties.pubsub = True

        assert properties.jid is not None
        if properties.jid.is_full:
            self._log.warning("PubSub event from full jid %s", properties.jid)
            self._log.warning(stanza)
            return

        event = stanza.getTag("event", namespace=Namespace.PUBSUB_EVENT)

        delete = event.getTag("delete")
        if delete is not None:
            node = delete.getAttr("node")
            properties.pubsub_event = PubSubEventData(node, deleted=True)
            return

        purge = event.getTag("purge")
        if purge is not None:
            node = purge.getAttr("node")
            properties.pubsub_event = PubSubEventData(node, purged=True)
            return

        items = event.getTag("items")
        if items is not None:
            node = items.getAttr("node")

            retract = items.getTag("retract")
            if retract is not None:
                id_ = retract.getAttr("id")
                properties.pubsub_event = PubSubEventData(node, id_, retracted=True)
                return

            if len(items.getChildren()) != 1:
                self._log.warning("PubSub event with != 1 item")
                self._log.warning(stanza)
                return

            item = items.getTag("item")
            if item is None:
                self._log.warning("No item node found")
                self._log.warning(stanza)
                return
            id_ = item.getAttr("id")
            properties.pubsub_event = PubSubEventData(node, id_, item)

    @iq_request_task
    def request_item(self, node: str, id_: str, jid: JID | None = None):
        task = yield

        response = yield _make_pubsub_request(node, id_=id_, jid=jid)

        if response.isError():
            raise PubSubStanzaError(response)

        item = _get_pubsub_item(response, node, id_)
        yield task.set_result(item)

    @iq_request_task
    def request_items(
        self, node: str, max_items: int | None = None, jid: JID | None = None
    ):
        _task = yield

        response = yield _make_pubsub_request(node, max_items=max_items, jid=jid)

        if response.isError():
            raise PubSubStanzaError(response)

        yield _get_pubsub_items(response, node)

    @iq_request_task
    def publish(
        self,
        node: str,
        item: Node,
        id_: str | None = None,
        options: dict[str, str] | None = None,
        jid: JID | None = None,
        force_node_options: bool = False,
    ):

        _task = yield

        request = _make_publish_request(node, item, id_, options, jid)
        response = yield request

        if response.isError():
            error = PubSubStanzaError(response)
            if not force_node_options or error.app_condition != "precondition-not-met":
                raise error

            result = yield self.reconfigure_node(node, options, jid)
            if is_error(result):
                raise result

            response = yield request
            if response.isError():
                raise PubSubStanzaError(response)

        jid = response.getFrom()
        item_id = _get_published_item_id(response, node, id_)
        yield PubSubPublishResult(jid, node, item_id)

    @iq_request_task
    def get_access_model(self, node: str):
        _task = yield

        result = yield self.get_node_configuration(node)

        raise_if_error(result)

        yield result.form["pubsub#access_model"].value

    @iq_request_task
    def set_access_model(self, node: str, model: Literal["open", "presence"]):
        task = yield

        if model not in ("open", "presence"):
            raise ValueError("Invalid access model")

        result = yield self.get_node_configuration(node)

        raise_if_error(result)

        try:
            access_model = result.form["pubsub#access_model"].value
        except Exception:
            yield task.set_error("warning", condition="access-model-not-supported")

        if access_model == model:
            jid = self._client.get_bound_jid().new_as_bare()
            yield CommonResult(jid=jid)

        result.form["pubsub#access_model"].value = model

        result = yield self.set_node_configuration(node, result.form)

        yield finalize(task, result)

    @iq_request_task
    def retract(self, node: str, id_: str, jid: JID | None = None, notify: bool = True):
        _task = yield

        response = yield _make_retract_request(node, id_, jid, notify)
        yield process_response(response)

    @iq_request_task
    def purge(self, node: str, jid: JID | None = None):
        _task = yield

        response = yield _make_purge_request(node, jid)
        yield process_response(response)

    @iq_request_task
    def delete(self, node: str, jid: JID | None = None):
        _task = yield

        response = yield _make_delete_request(node, jid)
        yield process_response(response)

    @iq_request_task
    def reconfigure_node(self, node: str, options: dict[str, str], jid=None):
        _task = yield

        result = yield self.get_node_configuration(node, jid)
        if is_error(result):
            raise result

        _apply_options(result.form, options)
        result = yield self.set_node_configuration(node, result.form, jid)
        yield result

    @iq_request_task
    def set_node_configuration(self, node: Node, form, jid: JID | None = None):
        _task = yield

        response = yield _make_node_configuration(node, form, jid)
        yield process_response(response)

    @iq_request_task
    def get_node_configuration(self, node: str, jid: JID | None = None):
        _task = yield

        response = yield _make_node_configuration_request(node, jid)

        if response.isError():
            raise PubSubStanzaError(response)

        jid = response.getFrom()
        form = _get_configure_form(response, node)
        yield PubSubNodeConfigurationResult(jid=jid, node=node, form=form)

    @iq_request_task
    def get_subscriptions(self, jid: JID):
        _task = yield

        response = yield _make_get_subscription_request(jid)

        if response.isError():
            raise PubSubStanzaError(response)

        pubsub_node = response.getTag("pubsub", namespace=Namespace.PUBSUB)
        if pubsub_node is None:
            raise MalformedStanzaError("pubsub node missing", response)

        subscriptions = pubsub_node.getTag("subscriptions", namespace=Namespace.PUBSUB)
        if subscriptions is None:
            raise MalformedStanzaError("subscriptions node missing", response)

        result = []

        for item in subscriptions.getTags("subscription", namespace=Namespace.PUBSUB):
            try:
                sub = PubSubSubscription.from_response(item)
            except MalformedStanzaError as error:
                self._log.warning(error)
                self._log.warning(error.stanza)
                continue

            result.append(sub)

        yield result

    @iq_request_task
    def subscribe(self, node: str, jid: str | None = None):
        _task = yield

        own_jid = self._client.get_bound_jid().new_as_bare()
        response = yield _make_subscribe_request(node, own_jid, jid)

        if response.isError():
            raise PubSubStanzaError(response)

        item = _get_subscription_item(response)
        if item is None:
            raise MalformedStanzaError("subscription node missing", response)

        yield PubSubSubscription.from_response(item)

    @iq_request_task
    def unsubscribe(self, node: str, jid: str | None = None):
        _task = yield

        own_jid = self._client.get_bound_jid().new_as_bare()
        response = yield _make_unsubscribe_request(node, own_jid, jid)
        yield process_response(response)


def get_pubsub_request(
    jid: JID, node: str, id_: str | None = None, max_items: int | None = None
) -> Iq:
    query = Iq("get", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    items = pubsub.addChild("items", {"node": node})
    if max_items is not None:
        items.setAttr("max_items", max_items)
    if id_ is not None:
        items.addChild("item", {"id": id_})
    return query


def get_pubsub_item(stanza: Node) -> Node | None:
    pubsub_node = stanza.getTag("pubsub")
    items_node = pubsub_node.getTag("items")
    return items_node.getTag("item")


def get_pubsub_items(stanza: Node, node: str | None = None) -> list[Node] | None:
    pubsub_node = stanza.getTag("pubsub")
    items_node = pubsub_node.getTag("items")
    if node is not None and items_node.getAttr("node") != node:
        return None

    if items_node is not None:
        return items_node.getTags("item")
    return None


def get_publish_options(config: dict[str, str]) -> Node:
    options = Node(Namespace.DATA + " x", attrs={"type": "submit"})
    field = options.addChild("field", attrs={"var": "FORM_TYPE", "type": "hidden"})
    field.setTagData("value", Namespace.PUBSUB_PUBLISH_OPTIONS)

    for var, value in config.items():
        field = options.addChild("field", attrs={"var": var})
        field.setTagData("value", value)
    return options


def _get_pubsub_items(response: Iq, node: str) -> list[Node]:
    pubsub_node = response.getTag("pubsub", namespace=Namespace.PUBSUB)
    if pubsub_node is None:
        raise MalformedStanzaError("pubsub node missing", response)

    items_node = pubsub_node.getTag("items")
    if items_node is None:
        raise MalformedStanzaError("items node missing", response)

    if items_node.getAttr("node") != node:
        raise MalformedStanzaError("invalid node attr", response)

    return items_node.getTags("item")


def _get_pubsub_item(response: Iq, node: str, id_: str) -> Node | None:
    items = _get_pubsub_items(response, node)

    if len(items) > 1:
        raise MalformedStanzaError("multiple items found", response)

    if not items:
        return None

    item = items[0]
    if item.getAttr("id") != id_:
        raise MalformedStanzaError("invalid item id", response)

    return item


def _get_subscription_item(response: Iq) -> Node | None:
    pubsub_node = response.getTag("pubsub", namespace=Namespace.PUBSUB)
    if pubsub_node is None:
        raise MalformedStanzaError("pubsub node missing", response)

    return pubsub_node.getTag("subscription", namespace=Namespace.PUBSUB)


def _make_pubsub_request(
    node: str,
    id_: str | None = None,
    max_items: int | None = None,
    jid: JID | None = None,
) -> Iq:
    query = Iq("get", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    items = pubsub.addChild("items", {"node": node})
    if max_items is not None:
        items.setAttr("max_items", max_items)
    if id_ is not None:
        items.addChild("item", {"id": id_})
    return query


def _get_configure_form(response: Iq, node: str):
    pubsub = response.getTag("pubsub", namespace=Namespace.PUBSUB_OWNER)
    if pubsub is None:
        raise MalformedStanzaError("pubsub node missing", response)

    configure = pubsub.getTag("configure")
    if configure is None:
        raise MalformedStanzaError("configure node missing", response)

    if node != configure.getAttr("node"):
        raise MalformedStanzaError("invalid node attribute", response)

    forms = configure.getTags("x", namespace=Namespace.DATA)
    for form in forms:
        dataform = extend_form(node=form)
        form_type = dataform.vars.get("FORM_TYPE")
        if form_type is None or form_type.value != Namespace.PUBSUB_CONFIG:
            continue

        return dataform

    raise MalformedStanzaError("no valid form type found", response)


def _get_published_item_id(response: Iq, node: str, id_: str | None) -> str | None:
    pubsub = response.getTag("pubsub", namespace=Namespace.PUBSUB)
    if pubsub is None:
        # https://xmpp.org/extensions/xep-0060.html#publisher-publish-success
        # If the publish request did not include an ItemID,
        # the IQ-result SHOULD include an empty <item/> element
        # that specifies the ItemID of the published item.
        #
        # If the server did not add a payload we assume the item was
        # published with the id we requested
        return id_

    publish = pubsub.getTag("publish")
    if publish is None:
        raise MalformedStanzaError("publish node missing", response)

    if node != publish.getAttr("node"):
        raise MalformedStanzaError("invalid node attribute", response)

    item = publish.getTag("item")
    if item is None:
        raise MalformedStanzaError("item node missing", response)

    item_id = item.getAttr("id")
    if id_ is not None and item_id != id_:
        raise MalformedStanzaError("invalid item id", response)

    return item_id


def _make_publish_request(
    node: str,
    item: Node,
    id_: str | None,
    options: dict[str, str] | None,
    jid: JID | None,
) -> Iq:
    query = Iq("set", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    publish = pubsub.addChild("publish", {"node": node})
    attrs = {}
    if id_ is not None:
        attrs = {"id": id_}
    publish.addChild("item", attrs, [item])
    if options:
        publish = pubsub.addChild("publish-options")
        publish.addChild(node=_make_publish_options(options))
    return query


def _make_publish_options(options: dict[str, str]) -> Node:
    data = Node(Namespace.DATA + " x", attrs={"type": "submit"})
    field = data.addChild("field", attrs={"var": "FORM_TYPE", "type": "hidden"})
    field.setTagData("value", Namespace.PUBSUB_PUBLISH_OPTIONS)

    for var, value in options.items():
        field = data.addChild("field", attrs={"var": var})
        field.setTagData("value", value)
    return data


def _make_retract_request(node: str, id_: str, jid: JID | None, notify: bool) -> Iq:
    query = Iq("set", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    attrs = {"node": node}
    if notify:
        attrs["notify"] = "true"
    retract = pubsub.addChild("retract", attrs=attrs)
    retract.addChild("item", {"id": id_})
    return query


def _make_purge_request(node: str, jid: JID | None) -> Iq:
    query = Iq("set", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB_OWNER)

    pubsub.addChild("purge", attrs={"node": node})
    return query


def _make_delete_request(node: str, jid: JID | None) -> Iq:
    query = Iq("set", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB_OWNER)

    pubsub.addChild("delete", attrs={"node": node})
    return query


def _make_node_configuration(node: Node, form, jid: JID | None) -> Iq:
    query = Iq("set", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB_OWNER)
    configure = pubsub.addChild("configure", {"node": node})
    form.setAttr("type", "submit")
    configure.addChild(node=form)
    return query


def _make_node_configuration_request(node: str, jid: JID | None) -> Iq:
    query = Iq("get", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB_OWNER)
    pubsub.addChild("configure", {"node": node})
    return query


def _make_get_subscription_request(
    jid: JID | None = None,
) -> Iq:
    query = Iq("get", to=jid)
    pubsub = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    pubsub.addChild("subscriptions")
    return query


def _make_subscribe_request(node: str, own_jid: JID, jid: JID | None = None) -> Iq:
    query = Iq("set", to=jid)
    pb = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    pb.addChild("subscribe", {"node": node, "jid": str(own_jid)})
    return query


def _make_unsubscribe_request(node: str, own_jid: JID, jid: JID | None = None) -> Iq:
    query = Iq("set", to=jid)
    pb = query.addChild("pubsub", namespace=Namespace.PUBSUB)
    pb.addChild("unsubscribe", {"node": node, "jid": str(own_jid)})
    return query


def _apply_options(form, options: dict[str, str]) -> None:
    for var, value in options.items():
        try:
            field = form[var]
        except KeyError:
            pass
        else:
            field.value = value


class PubSubNodeConfigurationResult(NamedTuple):
    jid: JID
    node: str
    form: Any


class PubSubPublishResult(NamedTuple):
    jid: JID
    node: str
    id: str


@dataclass
class PubSubSubscription:
    node: str
    jid: JID
    subscription: str
    subid: str | None

    @classmethod
    def from_response(cls, sub_item: Iq) -> PubSubSubscription:
        node = sub_item.getAttr("node")
        if not node:
            raise MalformedStanzaError("node attr is missing", sub_item)

        jid = sub_item.getAttr("jid")
        if not jid:
            raise MalformedStanzaError("jid attr is missing", sub_item)

        try:
            jid = JID.from_string(jid)
        except Exception:
            raise MalformedStanzaError("jid is invalidvalid", sub_item)

        subscription = sub_item.getAttr("subscription")
        if not subscription:
            raise MalformedStanzaError("subscription attr is missing", sub_item)

        if subscription not in ("none", "pending", "subscribed", "unconfigured"):
            raise MalformedStanzaError(
                "unknown subscription value '%s'" % subscription, sub_item
            )

        subid = sub_item.getAttr("subid")

        return cls(node=node, jid=jid, subscription=subscription, subid=subid)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import Chatstate
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Chatstates(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_chatstate,
                ns=Namespace.CHATSTATES,
                priority=15,
            ),
        ]

    def _process_message_chatstate(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        try:
            chatstate = parse_chatstate(stanza)
        except ValueError as error:
            self._log.warning("Invalid chatstate: %s", error)
            self._log.warning(stanza)
            return

        if chatstate is None:
            return

        if properties.is_mam_message:
            return

        properties.chatstate = chatstate


def parse_chatstate(stanza: Message) -> Chatstate | None:
    children = stanza.getChildren()
    for child in children:
        if child.getNamespace() == Namespace.CHATSTATES:
            return Chatstate(child.getName())
    return None



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import Literal
from typing import TYPE_CHECKING

from nbxmpp.const import PresenceShow
from nbxmpp.const import PresenceType
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import log_calls
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import ERR_BAD_REQUEST
from nbxmpp.protocol import Error as ErrorStanza
from nbxmpp.protocol import JID
from nbxmpp.protocol import NodeProcessed
from nbxmpp.protocol import Presence
from nbxmpp.structs import PresenceProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import error_factory

if TYPE_CHECKING:
    from nbxmpp.client import Client


class BasePresence(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="presence", callback=self._process_presence_base, priority=10
            ),
        ]

    def _process_presence_base(
        self, _client: Client, stanza: Presence, properties: PresenceProperties
    ) -> None:
        properties.type = self._parse_type(stanza)
        properties.priority = self._parse_priority(stanza)
        properties.show = self._parse_show(stanza)
        properties.jid = stanza.getFrom()
        properties.id = stanza.getID()
        properties.status = stanza.getStatus()

        if properties.type.is_error:
            properties.error = error_factory(stanza)

        own_jid = self._client.get_bound_jid()
        properties.self_presence = own_jid == properties.jid
        properties.self_bare = properties.jid.bare_match(own_jid)

    def _parse_priority(self, stanza: Presence) -> int:
        priority = stanza.getPriority()
        if priority is None:
            return 0

        try:
            priority = int(priority)
        except Exception:
            self._log.warning("Invalid priority value: %s", priority)
            self._log.warning(stanza)
            return 0

        if priority not in range(-129, 128):
            self._log.warning("Invalid priority value: %s", priority)
            self._log.warning(stanza)
            return 0

        return priority

    def _parse_type(self, stanza: Presence) -> PresenceType:
        type_ = stanza.getType()
        try:
            return PresenceType(type_)
        except ValueError:
            self._log.warning("Presence with invalid type received")
            self._log.warning(stanza)
            self._client.send_stanza(ErrorStanza(stanza, ERR_BAD_REQUEST))
            raise NodeProcessed

    def _parse_show(self, stanza: Presence) -> PresenceShow:
        show = stanza.getShow()
        if show is None:
            return PresenceShow.ONLINE
        try:
            return PresenceShow(stanza.getShow())
        except ValueError:
            self._log.warning("Presence with invalid show")
            self._log.warning(stanza)
            return PresenceShow.ONLINE

    @log_calls
    def unsubscribe(self, jid: JID) -> None:
        self.send(jid=jid, typ="unsubscribe")

    @log_calls
    def unsubscribed(self, jid: JID) -> None:
        self.send(jid=jid, typ="unsubscribed")

    @log_calls
    def subscribed(self, jid: JID) -> None:
        self.send(jid=jid, typ="subscribed")

    @log_calls
    def subscribe(
        self, jid: JID, status: str | None = None, nick: str | None = None
    ) -> None:
        self.send(jid=jid, typ="subscribe", status=status, nick=nick)

    def send(
        self,
        jid: JID | None = None,
        typ: str | None = None,
        priority: int | None = None,
        show: Literal["chat", "away", "xa", "dnd"] | None = None,
        status: str | None = None,
        nick: str | None = None,
        caps: dict[str, str] | None = None,
        idle_time=None,
        signed=None,
        muc: bool = False,
        muc_history=None,
        muc_password: str | None = None,
        extend=None,
    ) -> None:

        if show is not None and show not in ("chat", "away", "xa", "dnd"):
            raise ValueError("Invalid show value: %s" % show)

        presence = Presence(jid, typ, priority, show, status)
        if nick is not None:
            nick_tag = presence.setTag("nick", namespace=Namespace.NICK)
            nick_tag.setData(nick)

        if idle_time is not None:
            idle_node = presence.setTag("idle", namespace=Namespace.IDLE)
            idle_node.setAttr("since", idle_time)

        if caps is not None and typ != "unavailable":
            presence.setTag("c", namespace=Namespace.CAPS, attrs=caps)

        if signed is not None:
            presence.setTag(Namespace.SIGNED + " x").setData(signed)

        if muc or muc_history is not None or muc_password is not None:
            muc_x = presence.setTag(Namespace.MUC + " x")
            if muc_history is not None:
                muc_x.setTag("history", muc_history)

            if muc_password is not None:
                muc_x.setTagData("password", muc_password)

        if extend is not None:
            for node in extend:
                presence.addChild(node=node)

        self._client.send_stanza(presence)



from __future__ import annotations

from typing import Any

import logging
import operator
from dataclasses import dataclass

from nbxmpp.exceptions import FallbackLanguageError
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message


@dataclass(frozen=True)
class FallbackRange:
    start: int
    end: int


FallbackLangMapT = dict[str | None, FallbackRange | None]
FallbacksForT = dict[str, FallbackLangMapT | None]


def parse_fallback_indication(
    log: logging.Logger | logging.LoggerAdapter[Any],
    stanza: Message,
) -> FallbacksForT | None:

    fallbacks = stanza.getTags(
        "fallback",
        namespace=Namespace.FALLBACK,
    )
    if not fallbacks:
        return None

    fallbacks_for: FallbacksForT = {}
    for fallback in fallbacks:
        for_ = fallback.getAttr("for")
        if not for_:
            log.warning('Missing "for" attribute on fallback indication')
            return

        bodies = fallback.getTags("body")
        if not bodies:
            # No <body> means fallback is for all bodies
            fallbacks_for[for_] = None
            continue

        fallback_lang_map: FallbackLangMapT = {}
        for body in bodies:
            lang = body.getAttr("xml:lang") or None
            start = body.getAttr("start") or None
            end = body.getAttr("end") or None

            if type(start) is not type(end):
                log.warning("Incorrect range on fallback indication")
                return

            range_ = None
            if start is not None:
                assert end is not None
                try:
                    range_ = FallbackRange(start=int(start), end=int(end))
                except Exception:
                    log.warning("Incorrect range on fallback indication")
                    return

                if range_.start < 0 or range_.end < 0:
                    log.warning("Fallback range with negative numbers: %s", range_)
                    return

            fallback_lang_map[lang] = range_

        if fallback_lang_map:
            # Only store data for the key if there was at least one
            # valid body
            fallbacks_for[for_] = fallback_lang_map

    return fallbacks_for


def strip_fallback(
    fallbacks_for: FallbacksForT,
    fallback_ns: set[str],
    lang: str | None,
    text: str,
) -> str:
    fallbacks: list[FallbackRange] = []
    # Gather all fallbacks we support
    for ns in fallback_ns:
        try:
            fallback_lang_map = fallbacks_for[ns]
        except KeyError:
            continue

        if fallback_lang_map is None:
            # At least one of the fallbacks declared the whole
            # body as fallback
            return ""

        try:
            range_ = fallback_lang_map[lang]
        except KeyError:
            # Fallback missing for this language, assume the stanza is
            # malformed and return the whole text
            raise FallbackLanguageError

        if range_ is None:
            # No range means the whole body is considered a fallback
            return ""

        if range_.end > len(text):
            # Abort processing, because range is invalid
            return text

        fallbacks.append(range_)

    if not fallbacks:
        return text

    fallbacks.sort(key=operator.attrgetter("start"), reverse=True)

    stripped_text = text
    for range_ in fallbacks:
        stripped_text = stripped_text[: range_.start] + stripped_text[range_.end :]

    return stripped_text



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import MOODS
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import Node
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import MoodData
from nbxmpp.structs import StanzaHandler
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Mood(BaseModule):

    _depends = {"publish": "PubSub"}

    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_pubsub_mood,
                ns=Namespace.PUBSUB_EVENT,
                priority=16,
            ),
        ]

    def _process_pubsub_mood(
        self, _client: Client, stanza: Message, properties: MessageProperties
    ) -> None:
        if not properties.is_pubsub_event:
            return

        if properties.pubsub_event.node != Namespace.MOOD:
            return

        item = properties.pubsub_event.item
        if item is None:
            # Retract, Deleted or Purged
            return

        mood_node = item.getTag("mood", namespace=Namespace.MOOD)
        if not mood_node.getChildren():
            self._log.info("Received mood: %s - removed mood", properties.jid)
            return

        mood, text = None, None
        for child in mood_node.getChildren():
            name = child.getName().strip()
            if name == "text":
                text = child.getData()
            elif name in MOODS:
                mood = name

        if mood is None and mood_node.getChildren():
            self._log.warning("No valid mood value found")
            self._log.warning(stanza)
            raise NodeProcessed

        data = MoodData(mood, text)
        pubsub_event = properties.pubsub_event._replace(data=data)
        self._log.info("Received mood: %s - %s", properties.jid, data)

        properties.pubsub_event = pubsub_event

    @iq_request_task
    def set_mood(self, data: MoodData):
        task = yield

        item = Node("mood", {"xmlns": Namespace.MOOD})
        if data is not None and data.mood:
            item.addChild(data.mood)

            if data.text:
                item.addChild("text", payload=data.text)

        result = yield self.publish(Namespace.MOOD, item, id_="current")

        yield finalize(task, result)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import HTTPUploadStanzaError
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.structs import HTTPUploadData
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

ALLOWED_HEADERS = ["Authorization", "Cookie", "Expires"]


class HTTPUpload(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_slot(self, jid: JID, filename: str, size: int, content_type: str):
        _task = yield

        response = yield _make_request(jid, filename, size, content_type)
        if response.isError():
            raise HTTPUploadStanzaError(response)

        slot = response.getTag("slot", namespace=Namespace.HTTPUPLOAD_0)
        if slot is None:
            raise MalformedStanzaError("slot node missing", response)

        put_uri = slot.getTagAttr("put", "url")
        if put_uri is None:
            raise MalformedStanzaError("put uri missing", response)

        get_uri = slot.getTagAttr("get", "url")
        if get_uri is None:
            raise MalformedStanzaError("get uri missing", response)

        headers: dict[str, str] = {}
        for header in slot.getTag("put").getTags("header"):
            name = header.getAttr("name")
            if name not in ALLOWED_HEADERS:
                raise MalformedStanzaError(
                    "not allowed header found: %s" % name, response
                )

            data = header.getData()
            if "\n" in data:
                raise MalformedStanzaError("newline in header data found", response)

            headers[name] = data

        yield HTTPUploadData(put_uri=put_uri, get_uri=get_uri, headers=headers)


def _make_request(jid: JID, filename: str, size: int, content_type: str) -> Iq:
    iq = Iq(typ="get", to=jid)
    attr = {"filename": filename, "size": size, "content-type": content_type}
    iq.setTag(name="request", namespace=Namespace.HTTPUPLOAD_0, attrs=attr)
    return iq



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.const import AnonymityMode
from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import extend_form
from nbxmpp.modules.util import finalize
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Node
from nbxmpp.structs import MuclumbusItem
from nbxmpp.structs import MuclumbusResult
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client

# API Documentation
# https://search.jabber.network/docs/api


class Muclumbus(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def request_parameters(self, jid: str):
        task = yield

        response = yield _make_parameter_request(jid)
        if response.isError():
            raise StanzaError(response)

        search = response.getTag("search", namespace=Namespace.MUCLUMBUS)
        if search is None:
            raise MalformedStanzaError("search node missing", response)

        dataform = search.getTag("x", namespace=Namespace.DATA)
        if dataform is None:
            raise MalformedStanzaError("dataform node missing", response)

        self._log.info("Muclumbus parameters received")
        yield finalize(task, extend_form(node=dataform))

    @iq_request_task
    def set_search(
        self, jid: str, dataform, items_per_page: int = 50, after: str | None = None
    ):
        _task = yield

        response = yield _make_search_query(jid, dataform, items_per_page, after)
        if response.isError():
            raise StanzaError(response)

        result = response.getTag("result", namespace=Namespace.MUCLUMBUS)
        if result is None:
            raise MalformedStanzaError("result node missing", response)

        items = result.getTags("item")
        if not items:
            yield MuclumbusResult(first=None, last=None, max=None, end=True, items=[])

        set_ = result.getTag("set", namespace=Namespace.RSM)
        if set_ is None:
            raise MalformedStanzaError("set node missing", response)

        first = set_.getTagData("first")
        last = set_.getTagData("last")
        try:
            max_ = int(set_.getTagData("max"))
        except Exception:
            raise MalformedStanzaError("invalid max value", response)

        results: list[MuclumbusItem] = []
        for item in items:
            jid = item.getAttr("address")
            name = item.getTagData("name")
            nusers = item.getTagData("nusers")
            description = item.getTagData("description")
            language = item.getTagData("language")
            is_open = item.getTag("is-open") is not None

            try:
                anonymity_mode = AnonymityMode(item.getTagData("anonymity-mode"))
            except ValueError:
                anonymity_mode = AnonymityMode.UNKNOWN
            results.append(
                MuclumbusItem(
                    jid=jid,
                    name=name or "",
                    nusers=nusers or "",
                    description=description or "",
                    language=language or "",
                    is_open=is_open,
                    anonymity_mode=anonymity_mode,
                )
            )
        yield MuclumbusResult(
            first=first, last=last, max=max_, end=len(items) < max_, items=results
        )


def _make_parameter_request(jid: str) -> Iq:
    query = Iq(to=jid, typ="get")
    query.addChild(node=Node("search", attrs={"xmlns": Namespace.MUCLUMBUS}))
    return query


def _make_search_query(
    jid: str, dataform, items_per_page: int = 50, after: str | None = None
) -> Iq:
    search = Node("search", attrs={"xmlns": Namespace.MUCLUMBUS})
    search.addChild(node=dataform)
    rsm = search.addChild("set", namespace=Namespace.RSM)
    rsm.addChild("max").setData(items_per_page)
    if after is not None:
        rsm.addChild("after").setData(after)
    query = Iq(to=jid, typ="get")
    query.addChild(node=search)
    return query



# Copyright (C) 2020 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

import datetime as dt

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.dataforms import create_field
from nbxmpp.modules.dataforms import SimpleDataForm
from nbxmpp.modules.rsm import parse_rsm
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Node
from nbxmpp.structs import MAMPreferencesData
from nbxmpp.structs import MAMQueryData
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class MAM(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @iq_request_task
    def make_query(
        self,
        jid: JID,
        queryid: str | None = None,
        start: dt.datetime | None = None,
        end: dt.datetime | None = None,
        with_: str | None = None,
        after: str | None = None,
        max_: int = 70,
    ):

        _task = yield

        response = yield _make_request(jid, queryid, start, end, with_, after, max_)
        if response.isError():
            raise StanzaError(response)

        jid = response.getFrom()
        fin = response.getTag("fin", namespace=Namespace.MAM_2)
        if fin is None:
            raise MalformedStanzaError("fin node missing", response)

        rsm = parse_rsm(fin)
        if rsm is None:
            raise MalformedStanzaError("rsm set missing", response)

        complete = fin.getAttr("complete") == "true"
        if max_ != 0 and not complete:
            # max_ == 0 is a request for count of the items in a result set
            # in this case first and last will be absent
            # See: https://xmpp.org/extensions/xep-0059.html#count
            if rsm.first is None or rsm.last is None:
                raise MalformedStanzaError("first or last element missing", response)

        yield MAMQueryData(jid=jid, complete=complete, rsm=rsm)

    @iq_request_task
    def request_preferences(self):
        _task = yield

        response = yield _make_pref_request()
        if response.isError():
            raise StanzaError(response)

        prefs = response.getTag("prefs", namespace=Namespace.MAM_2)
        if prefs is None:
            raise MalformedStanzaError("prefs node missing", response)

        default = prefs.getAttr("default")
        if default is None:
            raise MalformedStanzaError("default attr missing", response)

        always_node = prefs.getTag("always")
        if always_node is None:
            raise MalformedStanzaError("always node missing", response)

        always = _get_preference_jids(always_node)

        never_node = prefs.getTag("never")
        if never_node is None:
            raise MalformedStanzaError("never node missing", response)

        never = _get_preference_jids(never_node)
        yield MAMPreferencesData(default=default, always=always, never=never)

    @iq_request_task
    def set_preferences(self, default: str | None, always: list[str], never: list[str]):
        _task = yield

        if default not in ("always", "never", "roster"):
            raise ValueError("Wrong default preferences type")

        response = yield _make_set_pref_request(default, always, never)
        yield process_response(response)


def _make_query_form(
    start: dt.datetime | None, end: dt.datetime | None, with_: str | None
) -> SimpleDataForm:
    fields = [create_field(typ="hidden", var="FORM_TYPE", value=Namespace.MAM_2)]

    if start:
        fields.append(
            create_field(
                typ="text-single",
                var="start",
                value=start.strftime("%Y-%m-%dT%H:%M:%SZ"),
            )
        )

    if end:
        fields.append(
            create_field(
                typ="text-single", var="end", value=end.strftime("%Y-%m-%dT%H:%M:%SZ")
            )
        )

    if with_:
        fields.append(create_field(typ="jid-single", var="with", value=with_))

    return SimpleDataForm(type_="submit", fields=fields)


def _make_rsm_query(max_: int | None, after: str | None) -> Node:
    rsm_set = Node("set", attrs={"xmlns": Namespace.RSM})
    if max_ is not None:
        rsm_set.setTagData("max", max_)
    if after is not None:
        rsm_set.setTagData("after", after)
    return rsm_set


def _make_request(
    jid: JID,
    queryid: str | None,
    start: dt.datetime | None,
    end: dt.datetime | None,
    with_: str | None,
    after: str | None,
    max_: int | None,
) -> Iq:
    iq = Iq(typ="set", to=jid, queryNS=Namespace.MAM_2)
    if queryid is not None:
        iq.getQuery().setAttr("queryid", queryid)

    payload = [_make_query_form(start, end, with_), _make_rsm_query(max_, after)]

    iq.setQueryPayload(payload)
    return iq


def _make_pref_request() -> Iq:
    iq = Iq("get", queryNS=Namespace.MAM_2)
    iq.setQuery("prefs")
    return iq


def _get_preference_jids(node: Iq) -> list[JID]:
    jids: list[JID] = []
    for item in node.getTags("jid"):
        jid = item.getData()
        if not jid:
            continue

        try:
            jid = JID.from_string(jid)
        except Exception:
            continue

        jids.append(jid)
    return jids


def _make_set_pref_request(default: str, always: list[str], never: list[str]) -> Iq:
    iq = Iq(typ="set")
    prefs = iq.addChild(
        name="prefs", namespace=Namespace.MAM_2, attrs={"default": default}
    )
    always_node = prefs.addChild(name="always")
    never_node = prefs.addChild(name="never")
    for jid in always:
        always_node.addChild(name="jid").setData(jid)

    for jid in never:
        never_node.addChild(name="jid").setData(jid)
    return iq



from __future__ import annotations

from typing import TYPE_CHECKING

import datetime as dt

from nbxmpp import Namespace
from nbxmpp import Node
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import RetractionData
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Retraction(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_message_retraction,
                ns=Namespace.MESSAGE_RETRACT_1,
                priority=20,
            ),
        ]

    def _process_message_retraction(
        self, _client: Client, stanza: Node, properties: MessageProperties
    ) -> None:

        retract, is_tombstone = _get_retract_element(stanza, properties)
        if retract is None:
            return

        retract_id = retract.getAttr("id")
        if not retract_id:
            self._log.warning("<retract> without id")
            self._log.warning(stanza)
            return

        timestamp = _parse_retraction_timestamp(retract, is_tombstone, properties)

        properties.retraction = RetractionData(
            id=retract_id, is_tombstone=is_tombstone, timestamp=timestamp
        )


def _get_retract_element(
    element: Node, properties: MessageProperties
) -> tuple[Node | None, bool]:
    """
    returns a tuple
        Node and a boolean value which signals if it is a tombstone
    """

    retract = element.getTag("retract", namespace=Namespace.MESSAGE_RETRACT_1)
    if retract is not None:
        return retract, False

    if not properties.is_mam_message:
        return None, False

    retracted = element.getTag("retracted", namespace=Namespace.MESSAGE_RETRACT_1)
    if retracted is not None:
        return retracted, True

    return None, False


def _parse_retraction_timestamp(
    retract: Node, is_tombstone: bool, properties: MessageProperties
) -> dt.datetime:

    if is_tombstone:
        stamp_attr = retract.getAttr("stamp")
        stamp = parse_datetime(stamp_attr, check_utc=True, convert="utc")
        if stamp is not None:
            assert isinstance(stamp, dt.datetime)
            return stamp

    if properties.is_mam_message:
        assert properties.mam is not None
        stamp = properties.mam.timestamp
    else:
        stamp = properties.timestamp

    return dt.datetime.fromtimestamp(stamp, dt.timezone.utc)



# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.modules.base import BaseModule
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import Message
from nbxmpp.structs import HTTPAuthData
from nbxmpp.structs import IqProperties
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

if TYPE_CHECKING:
    from nbxmpp.client import Client


class HTTPAuth(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)
        self._client = client
        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_http_auth,
                ns=Namespace.HTTP_AUTH,
                priority=40,
            ),
            StanzaHandler(
                name="iq",
                callback=self._process_http_auth,
                typ="get",
                ns=Namespace.HTTP_AUTH,
                priority=40,
            ),
        ]

    def _process_http_auth(
        self,
        _client: Client,
        stanza: Iq | Message,
        properties: IqProperties | MessageProperties,
    ) -> None:
        confirm = stanza.getTag("confirm", namespace=Namespace.HTTP_AUTH)
        if confirm is None:
            return

        attrs = confirm.getAttrs()
        body = stanza.getTagData("body")
        id_ = attrs.get("id")
        method = attrs.get("method")
        url = attrs.get("url")
        properties.http_auth = HTTPAuthData(id_, method, url, body)
        self._log.info("HTTPAuth received: %s %s %s %s", id_, method, url, body)



# Copyright (C) 2019 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of nbxmpp.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

from typing import TYPE_CHECKING

from nbxmpp.errors import MalformedStanzaError
from nbxmpp.errors import StanzaError
from nbxmpp.modules.base import BaseModule
from nbxmpp.modules.date_and_time import parse_datetime
from nbxmpp.modules.util import process_response
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.protocol import JID
from nbxmpp.protocol import Node
from nbxmpp.structs import AnnotationNote
from nbxmpp.task import iq_request_task

if TYPE_CHECKING:
    from nbxmpp.client import Client


class Annotations(BaseModule):
    def __init__(self, client: Client) -> None:
        BaseModule.__init__(self, client)

        self._client = client
        self.handlers = []

    @property
    def domain(self) -> str | None:
        return self._client.get_bound_jid().domain

    @iq_request_task
    def request_annotations(self):
        _task = yield

        response = yield _make_request()
        if response.isError():
            raise StanzaError(response)

        query = response.getQuery()
        storage = query.getTag("storage", namespace=Namespace.ROSTERNOTES)
        if storage is None:
            raise MalformedStanzaError("storage node missing", response)

        notes: list[AnnotationNote] = []
        for note in storage.getTags("note"):
            try:
                jid = JID.from_string(note.getAttr("jid"))
            except Exception as error:
                self._log.warning("Invalid JID: %s, %s", note.getAttr("jid"), error)
                continue

            cdate = note.getAttr("cdate")
            if cdate is not None:
                cdate = parse_datetime(cdate, epoch=True)

            mdate = note.getAttr("mdate")
            if mdate is not None:
                mdate = parse_datetime(mdate, epoch=True)

            data = note.getData()
            notes.append(AnnotationNote(jid=jid, cdate=cdate, mdate=mdate, data=data))

        self._log.info("Received annotations from %s:", self.domain)
        for note in notes:
            self._log.info(note)
        yield notes

    @iq_request_task
    def set_annotations(self, notes: list[AnnotationNote]):
        _task = yield

        for note in notes:
            self._log.info(note)

        response = yield _make_set_request(notes)
        yield process_response(response)


def _make_request() -> Iq:
    payload = Node("storage", attrs={"xmlns": Namespace.ROSTERNOTES})
    return Iq(typ="get", queryNS=Namespace.PRIVATE, payload=payload)


def _make_set_request(notes: list[AnnotationNote]) -> Iq:
    storage = Node("storage", attrs={"xmlns": Namespace.ROSTERNOTES})
    for note in notes:
        node = Node("note", attrs={"jid": note.jid})
        node.setData(note.data)
        if note.cdate is not None:
            node.setAttr("cdate", note.cdate)
        if note.mdate is not None:
            node.setAttr("mdate", note.mdate)
        storage.addChild(node=node)
    return Iq(typ="set", queryNS=Namespace.PRIVATE, payload=storage)
